// Hex Editor Control
// By Steven Lawrance

#include <windows.h>

typedef struct _HEXEDITOR_DATA {
  HGLOBAL *hData; // partitioned into nLines lines
  DWORD nLines;
  DWORD nBytesPerLine;
  DWORD nBytesTotal;
  DWORD curLine;  // which line is at the top
  DWORD curChar;
  DWORD maxLen;
  HFONT hFont;
  BOOL bHighlight;
} HEXEDITOR_DATA, *LPHEXEDITOR_DATA;

LRESULT CALLBACK HexEditorProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {

  case WM_CREATE :
	{
	  // Create a new window :-)!  Initialize data structure and set it
	  LPHEXEDITOR_DATA lpHexEditor = (LPHEXEDITOR_DATA) GlobalAlloc(GPTR, sizeof(HEXEDITOR_DATA);
	  SetWindowLong(hWnd, 0, (LONG) lpHexEditor);

	  // Set the initial scroll information
	  ShowScrollBar(hWnd, SB_VERT, TRUE);
	  SetScrollRange(hWnd, SB_VERT, 0, 0, TRUE);
	  return 0;
	}
  
  case WM_PAINT :
	{
	  // Paint the surface area
	  LPHEXEDITOR_DATA lpHexEditor = (LPHEXEDITOR_DATA) GetWindowLong(hWnd, 0);
	  TEXTMETRIC textMetrics;
	  LPTSTR displayStr = (LPTSTR) GlobalAlloc(GPTR, lpHexEditor->nBytesPerLine * 3 + 1);
	  LPTSTR lineStr;
	  RECT updateRect;
	  HDC hDC = (HDC) wParam;
	  DWORD i, t, bottom;
	  HFONT hOldFont;

	  // Get device information
	  hOldFont = SelectObject(hDC, lpHexEditor->hFont);
	  GetUpdateRect(hWnd, &updateRect, FALSE);
	  GetTextMetrics(hDC, &textMetrics);
	  bottom = updateRect.bottom / textMetrics.tmHeight + lpHexEditor->curLine;
	  if (updateRect.bottom % textMetrics.tmHeight != 0)
		bottom++;
	  if (bottom > lpHexEditor->nLines)
		bottom = lpHexEditor->nLines;

	  // Draw each line that is visible
	  for (i = updateRect.top / textMetrics.tmHeight + lpHexEditor->curLine; i < bottom; i++) {
		// Write the Hex part of the line
		lineStr = (LPTSTR) GlobalLock(lpHexEditor->hData[i]);
		for (t = 0; t < lpHexEditor->nBytesPerLine; t++)
		  wsprintf(displayStr, "%02X ", (UINT) lineStr[t]);
		lstrcat(displayStr, " ");
		TextOut(hDC, 0, (i - lpHexEditor->curLine) * textMetrics.tmHeight, displayStr, lpHexEditor->nBytesPerLine * 3 + 1);

		// Write the ASCII part of the line
		for (t = 0; t < lpHexEditor->nBytesPerLine; t++)
		  displayStr[t] = (lineStr[t] < 32)? '.': lineStr[t];
		displayStr[lpHexEditor->nBytesPerLine] = 0;
		TextOut(hDC, lpHexEditor->nBytesPerLine * textMetrics.tmMaxCharWidth * 3 + textMetrics.tmMaxCharWidth, (i - lpHexEditor->curLine) * textMetrics.tmHeight, displayStr, lpHexEditor->nBytesPerLine);

		// Unlock the current line's data
		GlobalUnlock(lpHexEditor->hData[i]);
	  }

	  // Unselect the font and cleanup
	  GlboalFree(GlobalHandle(displayStr));
	  SelectObject(hDC, hOldFont);
	  ValidateRect(hWnd, NULL);
	  return 0;
	}

  case WM_VSCROLL :
	{
	  // Scroll event
	  LPHEXEDITOR_DATA lpHexEditor = (LPHEXEDITOR_DATA) GetWindowLong(hWnd, 0);
	  TEXTMETRIC textMetrics;
	  RECT clientRect;
	  HFONT hOldFont;
	  WORD pos = HIWORD(wParam);
	  HDC hDC = GetDC(hWnd);

	  // Get the font and window extent information
	  hOldFont = SelectObject(hDC, lpHexEditor->hFont);
	  GetClientRect(hWnd, &clientRect);
	  GetTextMetrics(hDC, &textMetrics);
	  SelectObject(hDC, hOldFont);
	  ReleaseDC(hWnd, hDC);

	  // Handle the case appropriately
	  switch (LOWORD(wParam)) {

	  case SB_TOP :
		// Top of view
		lpHexEditor->curLine = 0;
		break;

	  case SB_BOTTOM :
		// Bottom of view
		if (clientRect.bottom / textMetrics.tmHeight >= lpHexEditor->nLines)
		  lpHexEditor->curLine = 0;
		else
		  lpHexEditor->curLine = lpHexEditor->nLines - clientRect.bottom / textMetrics.tmHeight;
		break;

      case SB_LINEDOWN :
		// Line down
		if (lpHexEditor->curLine + 1 < lpHexEditor->nLines)
		  lpHexEditor->curLine++;
		break;
 
	  case SB_LINEUP :
		// Line up
		if (lpHexEditor->curLine > 0)
		  lpHexEditor->curLine--;
		break;

	  case SB_PAGEDOWN :
		// Page down
		if (lpHexEditor->curLine + clientRect.bottom / textMetrics.tmHeight - 1 < lpHexEditor->nLines)
		  lpHexEditor->curLine += clientRect.bottom / textMetrics.tmHeight - 1;
		break;

	  case SB_PAGEUP :
		// Page up
		if (lpHexEditor->curLine - clientRect.bottom / textMetrics.tmHeight + 1 > 0)
		  lpHexEditor->curLine -= clientRect.bottom / textMetrics.tmHeight - 1;
		else
		  lpHexEditor->curLine = 0;
		break;
 
	  case SB_THUMBTRACK :
		// Thumb track (manually setting position)
		lpHexEditor->curLine = pos;
		break;

	  } // switch (LOWORD(wParam))

	  // Repaint the window
	  InvalidateRect(hWnd, NULL, FALSE);
	  return 0;
	}

  case WM_CHAR :
	{
	  // User typed in something
	  LPHEXEDITOR_DATA lpHexEditor = (LPHEXEDITOR_DATA) GetWindowLong(hWnd, 0);
	  DWORD key = wParam;

	  // Process the arrows
	  if (GetKeyState(VK_DOWN) < 0) {
		// Down arrow, move the selection to the next line
		if (lpHexEditor->curChar + lpHexEditor->nBytesPerLine < lpHexEditor->nBytesTotal)
		  lpHexEditor->curChar += lpHexEditor->nBytesPerLine;
		else
		  lpHexEditor->curChar = lpHexEditor->nBytesTotal - 1;
	  }
	  if (GetKeyState(VK_UP) < 0) {
		// Up arrow, move the selection to the next line
		if (lpHexEditor->curChar - lpHexEditor->nBytesPerLine > 0)
		  lpHexEditor->curChar -= lpHexEditor->nBytesPerLine;
		else
		  lpHexEditor->curChar = 0;
	  }

	  // Let the default window proc handle any keys we don't handle by not returning yet
	  break;
	}

  case HM_SETDATA :
	{
	  // Set the data for the control
	  LPHEXEDITOR_DATA lpHexEditor = (LPHEXEDITOR_DATA) GetWindowLong(hWnd, 0);

	  // Go through and 
	  

	  return 0;
	}
  }
  
  return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

BOOL RegisterHexEditorControl(HINSTANCE hInstance) {
  // Register the Hex Editor Custom Control
  WNDCLASS wc;

  // Set the window class identifiers
  wc.style = CS_DBLCLKS | CS_HREDRAW | CS_OWNDC | CS_VREDRAW;
  wc.lpfnWndProc = (WNDPROC) HexEditorProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = sizeof(LPHEXEDITOR_DATA);
  wc.hInstance = hInstance;
  wc.hIcon = IDI_APPLICATION;
  wc.hCursor = IDC_IBEAM;
  wc.hbrBackground = COLOR_BACKGROUND;
  wc.lpszMenuName = NULL
  wc.lpszClassName = "HexEdit";
  return RegisterClass(&wc); 
}
