// X10 CP290 Dynamic Link Library

// Purpose: To control the X10 CP290 interface by acting as a bridge between the
// 			program and the CP290 unit.

// Exported Functions:
//   CP290_GetBaseHouseCode
//   CP290_SetBaseHouseCode
//   CP290_GetTimerEvent
//   CP290_SetTimerEvent
//   CP290_GetTime
//   CP290_SetTime
//   CP290_GetIcon
//   CP290_SetIcon
//   CP290_Control
//   CP290_Initialize
//   CP290_Uninitialize
//   CP290_GetLastError
//   CP290_Reset

#include <windows.h>


// Global Variables
HANDLE ghComm;
DWORD gLastError;


// Constant Declarations
#define CP290_ON        0
#define CP290_OFF       1
#define CP290_DIM       2
#define CP290_NORMAL    0
#define CP290_SECURITY  1
#define CP290_TODAY     2
#define CP290_TOMORROW  3
#define CP290_CLEAR     4
#define CP290_SUNDAY    0
#define CP290_MONDAY    1
#define CP290_TUESDAY   2
#define CP290_WEDNESDAY 3
#define CP290_THURSDAY  4
#define CP290_FRIDAY    5
#define CP290_SATURDAY  6

// Error Message Constants
#define CP290ERR_WRITE          1
#define CP290ERR_READ           2
#define CP290ERR_CHECKSUM       3
#define CP290ERR_NO_ACK         4
#define CP290ERR_PORT_OPEN      5
#define CP290ERR_COMM_DCB       6
#define CP290ERR_COMM_STATE     7
#define CP290ERR_BAD_DAY        8
#define CP290ERR_BAD_HOUR       9
#define CP290ERR_BAD_MINUTE     10
#define CP290ERR_BAD_ID         11
#define CP290ERR_BAD_HOUSECODE  12
#define CP290ERR_BAD_COMMAND    13
#define CP290ERR_BAD_MODE       14
#define CP290ERR_BAD_DIMLEVEL   15
#define CP290ERR_BAD_DATABYTE   16
#define CP290ERR_MESSAGELENGTH  50


// Type Declarations
typedef struct _CP290ICON
{
  BYTE IconId;
  BYTE Data[2];
  BOOL InUse;
} CP290ICON;
typedef struct _CP290COMMAND
{
  int DimLevel;
  int Command;
  char HouseCode;
  BOOL Unit[16];
} CP290COMMAND;
typedef struct _CP290TIMEREVENT
{
  BYTE EventId;
  int Mode;
  SYSTEMTIME Time;
  BOOL Unit[16];
  BOOL Day[7];
  char HouseCode;
  int DimLevel;
  int Command;
  BOOL InUse;
} CP290TIMEREVENT;


// Function Prototypes
char __declspec(dllexport) CP290_GetBaseHouseCode(void);
BOOL __declspec(dllexport) CP290_SetBaseHouseCode(char BaseHouseCode);
BOOL __declspec(dllexport) CP290_GetTimerEvents(CP290TIMEREVENT *TimerEvent, int NumEvents);
BOOL __declspec(dllexport) CP290_SetTimerEvent(CP290TIMEREVENT *TimerEvent);
BOOL __declspec(dllexport) CP290_GetTime(SYSTEMTIME *Time);
BOOL __declspec(dllexport) CP290_SetTime(SYSTEMTIME *Time);
BOOL __declspec(dllexport) CP290_GetIcons(CP290ICON *Icon, int NumIcons);
BOOL __declspec(dllexport) CP290_SetIcon(CP290ICON *Icon);
BOOL __declspec(dllexport) CP290_Control(CP290COMMAND *CmdStruct);
BOOL __declspec(dllexport) CP290_Initialize(LPCSTR DeviceName);
BOOL __declspec(dllexport) CP290_Uninitialize(void);
DWORD __declspec(dllexport) CP290_GetLastError(LPSTR MessageText);
BOOL __declspec(dllexport) CP290_Reset(void);
BOOL SendSyncAndRequest(BYTE Request);
BOOL CheckCP290Unit(void);
BOOL WaitForAck(void);
BOOL ReadByte(BYTE *Byte);
BOOL SendByte(BYTE Byte);
BOOL ExtractBit(BYTE Byte, int Bit);
BYTE FormulateByte(BOOL D7, BOOL D6, BOOL D5, BOOL D4, BOOL D3, BOOL D2, BOOL D1, BOOL D0);
BYTE GetHouseCodeByte(char HouseCode);
char GetHouseCodeFromByte(BYTE HouseCode);
void GetUnitCodeBytes(BOOL *Units, BYTE *Byte1, BYTE *Byte2);
void GetUnitCodeFromBytes(BOOL *Units, BYTE Byte1, BYTE Byte2);
BYTE GetDayByte(BOOL *Day);


// DLL Entry Point

BOOL WINAPI DllEntryPoint(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  return TRUE;
}


char __declspec(dllexport) CP290_GetBaseHouseCode(void)
{
  BYTE Minute, Hour, Day, HouseCode, CheckSumUnit, CheckSumComputer;
  char HouseCodeChar;

  // Clear the input buffer
  CheckCP290Unit();

  // Send sync and request 4
  if (!SendSyncAndRequest(4))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  // Get the minute (not used in this function)
  if (!ReadByte(&Minute))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the hour (not used in this function)
  if (!ReadByte(&Hour))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the day (not used in this function)
  if (!ReadByte(&Day))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the house code
  if (!ReadByte(&HouseCode))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the checksum
  if (!ReadByte(&CheckSumUnit))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Calculate the checksum
  CheckSumComputer = (BYTE) (Minute + Hour + Day + HouseCode);

  // Compare the checksums
  if (CheckSumComputer != CheckSumUnit)
	 { gLastError = CP290ERR_CHECKSUM; return FALSE; }

  // Get the house code
  if ((HouseCodeChar = GetHouseCodeFromByte(HouseCode)) == 127)
	 { gLastError = CP290ERR_BAD_HOUSECODE; return FALSE; }

  // Success!  Return the house code
  return HouseCodeChar;
}


BOOL __declspec(dllexport) CP290_SetBaseHouseCode(char BaseHouseCode)
{
  // Set the base housecode
  BYTE HouseCode;

  // Convert the house code character into a byte that the unit understands
  HouseCode = GetHouseCodeByte(BaseHouseCode);

  // Clear the input buffer
  CheckCP290Unit();

  // Send sync and request zero
  if (!SendSyncAndRequest(0))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send the byte
  if (!SendByte(HouseCode))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  return TRUE;
}


BOOL __declspec(dllexport) CP290_GetTimerEvents(CP290TIMEREVENT *TimerEvent, int NumEvents)
{
  // This function reads every timer event from the unit into the specified
  // array for the first NumEvents elements
  BYTE ChecksumUnit, ChecksumComputer;
  BYTE Data[1024];
  int lcv, ReadLcv;

  // Initialize array to 255
  FillMemory(Data, 1024, 255);

  // Clear the input buffer
  CheckCP290Unit();

  // Send Sync and request 5
  if (!SendSyncAndRequest(5))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  // Read the bytes into the array
  lcv = 0;
  do
  {
	 if (!ReadByte(&Data[lcv]))
		{ gLastError = CP290ERR_READ; return FALSE; }
	 if (Data[lcv] < 255)
	 {
		for (ReadLcv = 1; ReadLcv <= 7; ReadLcv++)
		{
		  if (!ReadByte(&Data[lcv + ReadLcv]))
			 { gLastError = CP290ERR_READ; return FALSE; }
		}
	 }
//	 else
//		{ gLastError = CP290ERR_MESSAGELENGTH; return FALSE; }
	 lcv += 8;
  } while (lcv <= 1016);

  // Read the ChecksumUnit
  if (!ReadByte(&ChecksumUnit))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Calculate ChecksumComputer
  ChecksumComputer = 0;
  for (lcv = 0; lcv <= 127; lcv++)
	 if (Data[lcv * 8] != 255)
		for (ReadLcv = 0; ReadLcv <= 7; ReadLcv++)
		  ChecksumComputer += (BYTE) (Data[lcv * 8 + ReadLcv]);

  // Compare ChecksumComputer to ChecksumUnit
  if (ChecksumComputer != ChecksumUnit)
	 { gLastError = CP290ERR_CHECKSUM; return FALSE; }

  // Assign the events to the given array; indicate their Id, mode, day(s),
  // hour, minute, unit code(s), housecode, dim level, function, and usage flag.
  for (lcv = 0; (lcv <= 127) && (lcv <= NumEvents - 1); lcv++)
  {
	 TimerEvent[lcv].EventId = (BYTE) (lcv);
	 if (Data[lcv * 8] == 255)
	 {
		// Set this event to zero values and set the usage flag to FALSE
		TimerEvent[lcv].Mode = 0;
		TimerEvent[lcv].Time.wHour = 0;
		TimerEvent[lcv].Time.wMinute = 0;
		TimerEvent[lcv].Time.wDayOfWeek = 8;
		ZeroMemory(TimerEvent[lcv].Unit, 16 * sizeof(BOOL));
		FillMemory(TimerEvent[lcv].Day, 7 * sizeof(BOOL), TRUE);
		TimerEvent[lcv].HouseCode = 'A';
		TimerEvent[lcv].DimLevel = 0;
		TimerEvent[lcv].Command = 0;
		TimerEvent[lcv].InUse = FALSE;
	 }
	 if (Data[lcv * 8] < 255)
	 {
		int DayLcv, ModeLcv, CommandLcv;
		BYTE Mode, Command;

		// Set the mode
		TimerEvent[lcv].Mode = CP290_CLEAR + 1;
      Mode = 255;
		for (ModeLcv = CP290_NORMAL; (ModeLcv <= CP290_CLEAR) && (TimerEvent[lcv].Mode == CP290_CLEAR + 1); ModeLcv++)
		{
		  switch (ModeLcv)
		  {
			 case CP290_NORMAL   : Mode = FormulateByte(0, 0, 0, 0, 1, 0, 0, 0); break;
			 case CP290_SECURITY : Mode = FormulateByte(0, 0, 0, 0, 1, 0, 0, 1); break;
			 case CP290_TODAY    : Mode = FormulateByte(0, 0, 0, 0, 0, 1, 0, 0); break;
			 case CP290_TOMORROW : Mode = FormulateByte(0, 0, 0, 0, 0, 0, 1, 0); break;
			 case CP290_CLEAR    : Mode = FormulateByte(0, 0, 0, 0, 0, 0, 0, 0); break;
		  }
		  if (Mode == Data[lcv * 8])
			 TimerEvent[lcv].Mode = ModeLcv;
		}
		if (TimerEvent[lcv].Mode == CP290_CLEAR + 1)
		  { gLastError = CP290ERR_BAD_MODE; return FALSE; }

		// Set the day of week by converting the unit's day into Windows's day
		TimerEvent[lcv].Time.wDayOfWeek = 8;
		for (DayLcv = CP290_SUNDAY; DayLcv <= CP290_SATURDAY; DayLcv++)
		  TimerEvent[lcv].Day[DayLcv] = FALSE;
		for (DayLcv = CP290_MONDAY; DayLcv <= CP290_SATURDAY; DayLcv++)
		{
		  if (ExtractBit(Data[lcv * 8 + 1], 9 - DayLcv))
			 TimerEvent[lcv].Day[DayLcv] = TRUE;
		}
		if (ExtractBit(Data[lcv * 8 + 1], 2))
		  TimerEvent[lcv].Day[CP290_SUNDAY] = TRUE;

		// Set the hour
		if (Data[lcv * 8 + 2] > 23)
		  { gLastError = CP290ERR_BAD_HOUR; return FALSE; }
		else
		  TimerEvent[lcv].Time.wHour = (WORD) (Data[lcv * 8 + 2]);

		// Set the minute
		if (Data[lcv * 8 + 3] > 59)
		  { gLastError = CP290ERR_BAD_MINUTE; return FALSE; }
		else
		  TimerEvent[lcv].Time.wMinute = (WORD) (Data[lcv * 8 + 3]);

		// Assign the unit codes from the 2 bitmaps
		GetUnitCodeFromBytes(TimerEvent[lcv].Unit, Data[lcv * 8 + 4], Data[lcv * 8 + 5]);

		// Set the housecode
		TimerEvent[lcv].HouseCode = GetHouseCodeFromByte(Data[lcv * 8 + 6]);
		if (TimerEvent[lcv].HouseCode == 127)
		  { gLastError = CP290ERR_BAD_HOUSECODE; return FALSE; }

		// Set the level
		TimerEvent[lcv].DimLevel = (int) FormulateByte(0, 0, 0, 0, ExtractBit(Data[lcv * 8 + 7], 1), ExtractBit(Data[lcv * 8 + 7], 2), ExtractBit(Data[lcv * 8 + 7], 3), ExtractBit(Data[lcv * 8 + 7], 4));
		if ((TimerEvent[lcv].DimLevel < 0) || (TimerEvent[lcv].DimLevel > 15))
		  { gLastError = CP290ERR_BAD_DIMLEVEL; return FALSE; }

		TimerEvent[lcv].Command = CP290_DIM + 1;
		for (CommandLcv = CP290_ON; (CommandLcv <= CP290_DIM) && (TimerEvent[lcv].Command == CP290_DIM + 1); CommandLcv++)
		{
		  switch (CommandLcv)
		  {
			 case CP290_ON  : Command = FormulateByte(0, 0, 0, 0, 0, 0, 1, 0); break;
			 case CP290_OFF : Command = FormulateByte(0, 0, 0, 0, 0, 0, 1, 1); break;
			 case CP290_DIM : Command = FormulateByte(0, 0, 0, 0, 0, 1, 0, 1); break;
		  }
		  if ((Data[lcv * 8 + 7] - TimerEvent[lcv].DimLevel * 16) == Command)
			 TimerEvent[lcv].Command = CommandLcv;
		}
		if (TimerEvent[lcv].Command == CP290_DIM + 1)
		  { gLastError = CP290ERR_BAD_COMMAND; return FALSE; }

		// Set the usage flag to true
		TimerEvent[lcv].InUse = TRUE;
	 }
  }
  return TRUE;
}

BOOL __declspec(dllexport) CP290_SetTimerEvent(CP290TIMEREVENT *TimerEvent)
{
  // This function sets one timer event to the unit
  DWORD PrevMode;
  BYTE Checksum;
  BYTE Mode;
  BYTE Id[2];
  BYTE Unit[2];
  BYTE Day, Hour, Minute;
  BYTE HouseCode, Level, Command;

  // Formulate the Id bytes
  Id[0] = FormulateByte(ExtractBit(TimerEvent->EventId, 4), ExtractBit(TimerEvent->EventId, 5), ExtractBit(TimerEvent->EventId, 6), ExtractBit(TimerEvent->EventId, 7), ExtractBit(TimerEvent->EventId, 8), 0, 0, 0);
  Id[1] = FormulateByte(0, 0, 0, 0, 0, 0, ExtractBit(TimerEvent->EventId, 2), ExtractBit(TimerEvent->EventId, 3));

  // If the timer event is in use, then set the mode to clear
  PrevMode = TimerEvent->Mode;
  if (!TimerEvent->InUse)
	 TimerEvent->Mode = CP290_CLEAR;

  // Check and get the mode (normal, security, today, tomorrow, or clear)
  Mode = 255;
  switch (TimerEvent->Mode)
  {
	 case CP290_NORMAL   : Mode = FormulateByte(0, 0, 0, 0, 1, 0, 0, 0); break;
	 case CP290_SECURITY : Mode = FormulateByte(0, 0, 0, 0, 1, 0, 0, 1); break;
	 case CP290_TODAY    : Mode = FormulateByte(0, 0, 0, 0, 0, 1, 0, 0); break;
	 case CP290_TOMORROW : Mode = FormulateByte(0, 0, 0, 0, 0, 0, 1, 0); break;
	 case CP290_CLEAR    : Mode = FormulateByte(0, 0, 0, 0, 0, 0, 0, 0); break;
  }

  // Restore the original mode
  TimerEvent->Mode = PrevMode;

  // Check the validity of the mode
  if (Mode == 255)
	 { gLastError = CP290ERR_BAD_MODE; return FALSE; }

  // Get the day bitmap
  Day = GetDayByte(TimerEvent->Day);

  // Check and assign the hour and the minute
  Hour = (BYTE) (TimerEvent->Time.wHour);
  Minute = (BYTE) (TimerEvent->Time.wMinute);
  if (Hour > 23)
	 { gLastError = CP290ERR_BAD_HOUR; return FALSE; }
  if (Minute > 59)
	 { gLastError = CP290ERR_BAD_MINUTE; return FALSE; }

  // Get the unit code bitmaps
  GetUnitCodeBytes(TimerEvent->Unit, &Unit[1], &Unit[0]);

  // Check the house code and assign its byte
  if ((HouseCode = GetHouseCodeByte(TimerEvent->HouseCode)) == 255)
	 { gLastError = CP290ERR_BAD_HOUSECODE; return FALSE; }

  // Check the command type and create its byte
  Command = 0;
  switch (TimerEvent->Command)
  {
	 case CP290_ON  : Command = FormulateByte(0, 0, 0, 0, 0, 0, 1, 0); break;
	 case CP290_OFF : Command = FormulateByte(0, 0, 0, 0, 0, 0, 1, 1); break;
	 case CP290_DIM : Command = FormulateByte(0, 0, 0, 0, 0, 1, 0, 1); break;
  }
  if (Command == 0)
	 { gLastError = CP290ERR_BAD_COMMAND; return FALSE; }

  // Check the level and create its byte (force zero if event is not CP290_DIM)
  if ((TimerEvent->DimLevel <= 15) && (TimerEvent->DimLevel >= 0))
	if (TimerEvent->Command != CP290_DIM)
	  Level = 0;
	else
	  Level = (BYTE) (TimerEvent->DimLevel);
  else
	 { gLastError = CP290ERR_BAD_DIMLEVEL; return FALSE; }

  // Calculate the Checksum
  Checksum = (BYTE) (Mode + Day + Hour + Minute + Unit[0] + Unit[1] + HouseCode + (Level * 16 + Command));

  // Clear the input buffer
  CheckCP290Unit();

  // Send Sync and request 3
  if (!SendSyncAndRequest(3))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Id bytes (event number)
  if (!SendByte(Id[0]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte(Id[1]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Mode
  if (!SendByte(Mode))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Day(s), Hour, and Minute
  if (!SendByte(Day))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte(Hour))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte(Minute))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Unit code bitmap
  if (!SendByte(Unit[0]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte(Unit[1]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Housecode, Level, and Function
  if (!SendByte(HouseCode))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte((BYTE) (Level * 16 + Command)))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Checksum
  if (!SendByte(Checksum))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  return TRUE;
}


BOOL __declspec(dllexport) CP290_GetTime(SYSTEMTIME *Time)
{
  BYTE Minute, Hour, Day, HouseCode, CheckSumUnit, CheckSumComputer;
  BOOL DayArray[7];
  int lcv;

  // Clear the input buffer
  CheckCP290Unit();

  // Send sync and request 4
  if (!SendSyncAndRequest(4))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  // Get the minute
  if (!ReadByte(&Minute))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the hour
  if (!ReadByte(&Hour))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the day
  if (!ReadByte(&Day))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the house code (not used in this function)
  if (!ReadByte(&HouseCode))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Get the checksum
  if (!ReadByte(&CheckSumUnit))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Calculate the checksum
  CheckSumComputer = (BYTE) (Minute + Hour + Day + HouseCode);

  // Compare the checksums
  if (CheckSumComputer != CheckSumUnit)
	 { gLastError = CP290ERR_CHECKSUM; return FALSE; }

  // Assign Time its values
  Time->wMinute = Minute;
  Time->wHour = Hour;
  Time->wDayOfWeek = 8;
  for (lcv = CP290_SUNDAY; lcv <= CP290_SATURDAY; lcv++)
	 DayArray[lcv] = FALSE;
  for (lcv = CP290_SUNDAY; (lcv <= CP290_SATURDAY) && (Time->wDayOfWeek == 8); lcv++)
  {
	 DayArray[lcv] = TRUE;
	 if (Day == GetDayByte(DayArray))
		Time->wDayOfWeek = (WORD) lcv;
	 DayArray[lcv] = FALSE;
  }

  // Success!
  return TRUE;
}


BOOL __declspec(dllexport) CP290_SetTime(SYSTEMTIME *Time)
{
  BYTE Minute, Hour, Day, Checksum;
  int lcv;

  // Assign the minutes to a byte
  if (Time->wMinute > 59)
	 { gLastError = CP290ERR_BAD_MINUTE; return FALSE; }
  Minute = (BYTE) (Time->wMinute);

  // Assign the hours to a byte
  if (Time->wHour > 23)
	 { gLastError = CP290ERR_BAD_HOUR; return FALSE; }
  Hour = (BYTE) (Time->wHour);

  // Assign the day to a byte
  if (Time->wDayOfWeek > 6)
	 { gLastError = CP290ERR_BAD_DAY; return FALSE; }
  if (Time->wDayOfWeek == 0)
	 Day = 64;
  else
  {
	 Day = 1;
	 for (lcv = 2; lcv <= Time->wDayOfWeek; lcv++)
		Day = (BYTE) (Day * 2);
  }

  // Calculate the checksum
  Checksum = (BYTE) (Minute + Hour + Day);

  // Clear the input buffer
  CheckCP290Unit();

  // Send Sync and request 2
  if (!SendSyncAndRequest(2))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Minute
  if (!SendByte(Minute))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Hour
  if (!SendByte(Hour))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Day
  if (!SendByte(Day))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Checksum
  if (!SendByte(Checksum))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  return TRUE;
}

BOOL __declspec(dllexport) CP290_GetIcons(CP290ICON *Icon, int NumIcons)
{
  BYTE ChecksumUnit, ChecksumComputer;
  BYTE Data[512];
  int lcv;

  // Initialize array to 255
  FillMemory(Data, 512, 255);

  // Clear the input buffer
  CheckCP290Unit();

  // Send Sync and request 6
  if (!SendSyncAndRequest(6))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  // Read the bytes into the array
  lcv = 0;
  do
  {
	 if (!ReadByte(&Data[lcv]))
		{ gLastError = CP290ERR_READ; return FALSE; }
	 if (Data[lcv] != 255)
	 {
		if (!ReadByte(&Data[lcv + 1]))
		  { gLastError = CP290ERR_READ; return FALSE; }
	 }
	 lcv += 2;
  } while (lcv <= 510);

  // Read the ChecksumUnit
  if (!ReadByte(&ChecksumUnit))
	 { gLastError = CP290ERR_READ; return FALSE; }

  // Calculate ChecksumComputer
  ChecksumComputer = 0;
  for (lcv = 0; lcv <= 510; lcv += 2)
	 if (Data[lcv] != 255)
		ChecksumComputer += (BYTE) (Data[lcv] + Data[lcv + 1]);

  // Compare ChecksumComputer to ChecksumUnit
  if (ChecksumComputer != ChecksumUnit)
	 { gLastError = CP290ERR_CHECKSUM; return FALSE; }

  // Assign the icons to the given array; indicate their Id, their Data (if any), and
  // their usage flag
  for (lcv = 0; (lcv <= 255) && (lcv <= NumIcons - 1); lcv++)
  {
	 Icon[lcv].IconId = (BYTE) (lcv);
	 if (Data[lcv * 2] == 255)
	 {
		Icon[lcv].Data[0] = 0;
		Icon[lcv].Data[1] = 0;
		Icon[lcv].InUse = FALSE;
	 }
	 else
	 {
		Icon[lcv].Data[0] = Data[lcv * 2];
		Icon[lcv].Data[1] = Data[lcv * 2 + 1];
		Icon[lcv].InUse = TRUE;
	 }
  }
  return TRUE;
}

BOOL __declspec(dllexport) CP290_SetIcon(CP290ICON *Icon)
{
  BYTE Checksum;
  BYTE Id[2];

  // Formulate the Id bytes
  Id[0] = FormulateByte(ExtractBit(Icon->IconId, 2), ExtractBit(Icon->IconId, 3), ExtractBit(Icon->IconId, 4), ExtractBit(Icon->IconId, 5), ExtractBit(Icon->IconId, 6), ExtractBit(Icon->IconId, 7), ExtractBit(Icon->IconId, 8), 0);
  Id[1] = FormulateByte(0, 0, 0, 0, 0, 1, 0, ExtractBit(Icon->IconId, 1));

  // If the icon is not in use, set the data bytes accordingly
  if (!Icon->InUse)
  {
	 Icon->Data[0] = 0;
	 Icon->Data[1] = 0;
  }
  else
  {
	 // Since the byte was intended to be used, we must check the first icon byte
	 // for a careless error (it cannot be zero or 255)
	 if ((Icon->Data[0] == 0) || (Icon->Data[0] == 255))
		{ gLastError = CP290ERR_BAD_DATABYTE; return FALSE; }
  }

  // Calculate the Checksum
  Checksum = (BYTE) (Icon->Data[0] + Icon->Data[1]);

  // Clear the input buffer
  CheckCP290Unit();

  // Send Sync and request 3
  if (!SendSyncAndRequest(3))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Id bytes
  if (!SendByte(Id[0]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte(Id[1]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Icon Data
  if (!SendByte(Icon->Data[0]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte(Icon->Data[1]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Checksum
  if (!SendByte(Checksum))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  return TRUE;
}

BOOL __declspec(dllexport) CP290_Control(CP290COMMAND *CmdStruct)
{
  // Send a command to the CP290
  BYTE Command;
  BYTE Level;
  BYTE HouseCode;
  BYTE Unit[2];
  BYTE Checksum;
  BYTE CmdLevel;

  // Check the command type and create its byte
  Command = 0;
  switch (CmdStruct->Command)
  {
	 case CP290_ON  : Command = FormulateByte(0, 0, 0, 0, 0, 0, 1, 0); break;
	 case CP290_OFF : Command = FormulateByte(0, 0, 0, 0, 0, 0, 1, 1); break;
	 case CP290_DIM : Command = FormulateByte(0, 0, 0, 0, 0, 1, 0, 1); break;
  }
  if (Command == 0)
	 { gLastError = CP290ERR_BAD_COMMAND; return FALSE; }

  // Check the level and create its byte (force zero if command is not CP290_DIM)
  if ((CmdStruct->DimLevel <= 15) && (CmdStruct->DimLevel >= 0))
	if (CmdStruct->Command != CP290_DIM)
	  Level = 0;
	else
	  Level = (BYTE) CmdStruct->DimLevel;
  else
	 { gLastError = CP290ERR_BAD_DIMLEVEL; return FALSE; }
  CmdLevel = (BYTE) (Command + Level * 16);

  // Check the house code and assign its byte
  if ((HouseCode = GetHouseCodeByte(CmdStruct->HouseCode)) == 255)
	 { gLastError = CP290ERR_BAD_HOUSECODE; return FALSE; }

  // Get the unit code bytes from array
  GetUnitCodeBytes(CmdStruct->Unit, &Unit[0], &Unit[1]);

  // Calculate Checksum
  Checksum = (BYTE) (CmdLevel + HouseCode + Unit[0] + Unit[1]);

  // Clear the input buffer
  CheckCP290Unit();

  // Send Sync and request 1
  if (!SendSyncAndRequest(1))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Level and Function
  if (!SendByte(CmdLevel))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send House Code
  if (!SendByte(HouseCode))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Unit Bitmaps
  if (!SendByte(Unit[0]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  if (!SendByte(Unit[1]))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Send Checksum
  if (!SendByte(Checksum))
	 { gLastError = CP290ERR_WRITE; return FALSE; }

  // Wait for Acknowledgement for the command
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }

  // Wait for Acknowledgement for the unit to complete the command
  if (!WaitForAck())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }
  if (!CheckCP290Unit())
	 { gLastError = CP290ERR_NO_ACK; return FALSE; }
  return TRUE;
}

BOOL __declspec(dllexport) CP290_Initialize(LPCSTR DeviceName)
{
  // Open the communications port specified
  COMMTIMEOUTS cto;
  DCB dcb;
  if ((ghComm = CreateFile(DeviceName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE)
	 { gLastError = CP290ERR_PORT_OPEN; return FALSE; }

  // Set the communications buffers to 1024 bytes for the receive and 128 bytes
  // for the transmit
  SetupComm(ghComm, 1024, 128);

  // Set the timeouts for the reads to 12 seconds, 10 seconds between
  // characters, and 5 seconds for each write to the unit
  cto.ReadIntervalTimeout = 10000;
  cto.ReadTotalTimeoutMultiplier = 1;
  cto.ReadTotalTimeoutConstant = 12000;
  cto.WriteTotalTimeoutMultiplier = 1;
  cto.WriteTotalTimeoutConstant = 5000;
  SetCommTimeouts(ghComm, &cto);

  // Set the baud to 600, N, 8, 1
  GetCommState(ghComm, &dcb);
  if (!BuildCommDCB("baud=600 parity=N data=8 stop=1", &dcb))
  {
	 CloseHandle(ghComm);
	 { gLastError = CP290ERR_COMM_DCB; return FALSE; }
  }
  dcb.fParity = FALSE;
  dcb.fOutxCtsFlow = FALSE;
  dcb.fOutxDsrFlow = FALSE;
  dcb.fDtrControl = DTR_CONTROL_ENABLE;
  dcb.fDsrSensitivity = FALSE;
  dcb.fRtsControl = RTS_CONTROL_ENABLE;
  dcb.fAbortOnError = FALSE;
  if (!SetCommState(ghComm, &dcb))
  {
	 CloseHandle(ghComm);
	 { gLastError = CP290ERR_COMM_STATE; return FALSE; }
  }

  // Everything was ok; return true
  return TRUE;
}

BOOL __declspec(dllexport) CP290_Uninitialize(void)
{
  // Close the communications port
  return CloseHandle(ghComm);
}

DWORD __declspec(dllexport) CP290_GetLastError(LPSTR MessageText)
{
  DWORD LastError = gLastError;
  if (MessageText != NULL)
  {
	 switch (LastError)
	 {
		case 0  : lstrcpy(MessageText, "There was no error"); break;
		case 1  : lstrcpy(MessageText, "The port could not be written to"); break;
		case 2  : lstrcpy(MessageText, "The port could not be read from"); break;
		case 3  : lstrcpy(MessageText, "The unit had sent a bad checksum"); break;
		case 4  : lstrcpy(MessageText, "The unit was not responding"); break;
		case 5  : lstrcpy(MessageText, "The port could not be opened"); break;
		case 6  : lstrcpy(MessageText, "The BuildCommDcb function failed (internal)"); break;
		case 7  : lstrcpy(MessageText, "The SetCommState function failed (internal)"); break;
		case 8  : lstrcpy(MessageText, "The day was invalid"); break;
		case 9  : lstrcpy(MessageText, "The hour was invalid"); break;
		case 10 : lstrcpy(MessageText, "The minute was invalid"); break;
		case 11 : lstrcpy(MessageText, "The identification code was invalid"); break;
		case 12 : lstrcpy(MessageText, "The house code was invalid"); break;
		case 13 : lstrcpy(MessageText, "The command type was invalid"); break;
		case 14 : lstrcpy(MessageText, "The security mode was invalid"); break;
		case 15 : lstrcpy(MessageText, "The dim level was invalid"); break;
		case 16 : lstrcpy(MessageText, "The first data byte was either zero or 255"); break;
		case 50 : lstrcpy(MessageText, "The length of the packet was too short"); break;
	 }
  }
  gLastError = FALSE;
  return LastError;
}

BOOL __declspec(dllexport) CP290_Reset(void)
{
  // Clear the memory contents of the unit by sending Id 7 and wait for a
  // response after 12 seconds
  if (!SendSyncAndRequest(7))
	 { gLastError = CP290ERR_WRITE; return FALSE; }
  Sleep(12000);
  return TRUE;
}

BOOL SendSyncAndRequest(BYTE Request)
{
  // Send the Sync command and the request ID to the unit
  int lcv;
  FlushFileBuffers(ghComm);
  for (lcv = 1; lcv <= 16; lcv++)
  {
	 // Write 16 bytes of FF to the unit, leaving 1 ms between each byte
	 if (!SendByte(255))
		return FALSE;
  }

  // Send request
  if (!SendByte(Request))
	 return FALSE;
  return TRUE;
}

BOOL CheckCP290Unit(void)
{
  // Clear the input buffer; first, we will set the read timeout to 100 ms
  COMMTIMEOUTS ctoExisting, cto;
  BYTE NotUsed;
  GetCommTimeouts(ghComm, &ctoExisting);
  GetCommTimeouts(ghComm, &cto);
  cto.ReadIntervalTimeout = 10000;
  cto.ReadTotalTimeoutMultiplier = 1;
  cto.ReadTotalTimeoutConstant = 100;
  SetCommTimeouts(ghComm, &cto);
  while (ReadByte(&NotUsed));
  SetCommTimeouts(ghComm, &ctoExisting);
  return TRUE;
}

BOOL WaitForAck(void)
{
  // Wait for the CP290 to send an acknowledgement (6 FFs and 1 status byte)
  BYTE ReadBuffer[7];
  DWORD BytesRead;
  int lcv;
  ZeroMemory(ReadBuffer, sizeof(BYTE) * 7);
  BytesRead = 0;
  ReadFile(ghComm, ReadBuffer, 7, &BytesRead, NULL);
  if (BytesRead != 7)
	 return FALSE;
  for (lcv = 0; lcv <= 5; lcv++)
	 if (ReadBuffer[lcv] != 255)
		return FALSE;
  return TRUE;
}

BOOL ReadByte(BYTE *Byte)
{
  // Read a byte of data from the unit
  DWORD BytesRead;
  *Byte = 0;
  BytesRead = 0;
  ReadFile(ghComm, Byte, 1, &BytesRead, NULL);
  if (BytesRead < 1)
	 return FALSE;
  return TRUE;
}

BOOL SendByte(BYTE Byte)
{
  // SendByte sends a byte of data to the unit and waits for a millisecond
  DWORD BytesWritten;

  if (WriteFile(ghComm, &Byte, 1, &BytesWritten, NULL) == FALSE)
	 return FALSE;
  if (BytesWritten == 0)
	 return FALSE;
  Sleep(1);
  return TRUE;
}

BYTE FormulateByte(BOOL D7, BOOL D6, BOOL D5, BOOL D4, BOOL D3, BOOL D2, BOOL D1, BOOL D0)
{
  // Assign the 8 bit parameters to the byte and return the byte
  // Here is the order of the bit placement (from bit 0 to 7):
  // D7 D6 D5 D4 D3 D2 D1 D0

  BYTE Result;
  Result = (BYTE) D0;
  Result += (BYTE) (D1 * 2);
  Result += (BYTE) (D2 * 4);
  Result += (BYTE) (D3 * 8);
  Result += (BYTE) (D4 * 16);
  Result += (BYTE) (D5 * 32);
  Result += (BYTE) (D6 * 64);
  Result += (BYTE) (D7 * 128);
  return Result;
}

BOOL ExtractBit(BYTE Byte, int Bit)
{
  // ExtractBit returns the bit value of the Bit position in Byte
  Byte = (BYTE) (Byte << (Bit - 1));
  Byte = (BYTE) (Byte >> 7);
  return (BOOL) Byte;
}

BYTE GetHouseCodeByte(char HouseCode)
{
  BYTE hcb;

  // Determine the house code byte after changing the case to capital
  // If HouseCode is invalid, 255 is returned
  hcb = 255;
  if ((HouseCode <= 'p') && (HouseCode >= 'a'))
	 HouseCode -= (char) ('a' - 'A');
  switch (HouseCode)
  {
	 case 'A' : hcb = 96; break;
	 case 'B' : hcb = 224; break;
	 case 'C' : hcb = 32; break;
	 case 'D' : hcb = 160; break;
	 case 'E' : hcb = 16; break;
	 case 'F' : hcb = 144; break;
	 case 'G' : hcb = 80; break;
	 case 'H' : hcb = 208; break;
	 case 'I' : hcb = 112; break;
	 case 'J' : hcb = 240; break;
	 case 'K' : hcb = 48; break;
	 case 'L' : hcb = 176; break;
	 case 'M' : hcb = 0; break;
	 case 'N' : hcb = 128; break;
	 case 'O' : hcb = 64; break;
	 case 'P' : hcb = 192; break;
  }
  return hcb;
}

char GetHouseCodeFromByte(BYTE HouseCode)
{
  char hcb;

  // Assign hcb a meaningful house code
  hcb = 127;
  switch (HouseCode)
  {
	 case 96  : hcb = 'A'; break;
	 case 224 : hcb = 'B'; break;
	 case 32  : hcb = 'C'; break;
	 case 160 : hcb = 'D'; break;
	 case 16  : hcb = 'E'; break;
	 case 144 : hcb = 'F'; break;
	 case 80  : hcb = 'G'; break;
	 case 208 : hcb = 'H'; break;
	 case 112 : hcb = 'I'; break;
	 case 240 : hcb = 'J'; break;
	 case 48  : hcb = 'K'; break;
	 case 176 : hcb = 'L'; break;
	 case 0   : hcb = 'M'; break;
	 case 128 : hcb = 'N'; break;
	 case 64  : hcb = 'O'; break;
	 case 192 : hcb = 'P'; break;
  }
  return hcb;
}

void GetUnitCodeBytes(BOOL *Units, BYTE *Byte1, BYTE *Byte2)
{
  // Get the byte bitmaps from the Units array
  *Byte1 = FormulateByte(Units[8] ,Units[9], Units[10], Units[11], Units[12], Units[13], Units[14], Units[15]);
  *Byte2 = FormulateByte(Units[0] ,Units[1], Units[2], Units[3], Units[4], Units[5], Units[6], Units[7]);
}

void GetUnitCodeFromBytes(BOOL *Units, BYTE Byte1, BYTE Byte2)
{
  // Get the Units array from the byte bitmaps by assigning the true/false
  // value of each bit to their respective member in the Units array
  int lcv;
  for (lcv = 0; lcv <= 7; lcv++)
	 Units[lcv] = ExtractBit(Byte1, lcv + 1);
  for (lcv = 8; lcv <= 15; lcv++)
	 Units[lcv] = ExtractBit(Byte2, lcv - 7);
}

BYTE GetDayByte(BOOL *Day)
{
  // Get the day byte from the day array
  return FormulateByte(0, Day[CP290_SUNDAY], Day[CP290_SATURDAY], Day[CP290_FRIDAY], Day[CP290_THURSDAY], Day[CP290_WEDNESDAY], Day[CP290_TUESDAY], Day[CP290_MONDAY]);
}

